#include <stdio.h>
#include <string.h>
#include <time.h>
#include "common.h"
#include "hash.h"
#include "y.tab.h"
#include "iwrite.h"



#define PROG_VERSION "0.7-beta"
#ifndef DEFAULT_TEMPLATE_PATH
#define DEFAULT_TEMPLATE_PATH "/usr/local/lib/lddk/templates"
#endif
#ifndef DEFAULT_INCLUDE_PATH
#define DEFAULT_INCLUDE_PATH  ".:/usr/local/lib/lddk/ddllib"
#endif


#define DEVEL_PATH devel_path
#define TEMPLATE_PATH template_path


int driver_methods=0;
int driver_major = 0;
int driver_minor_to = 0;
int driver_minor = 0;

char devel_path[512];
char template_path[4096];
char include_path[4069];

DynString *driver_objects;
DynString *driver_sources;
DynString *driver_headers;
DynString *lib_objects;
DynString *lib_sources;
DynString *lib_headers;

DynString *doc_buffer = NULL;

DynString *submod_includes;
DynString *submod_cases;


IFILE *swigfile = NULL;
char CR_file[512];

/*==========================================================================*/

dump_CRNotice( FILE *outfile, int cflag ) {
    char iline[180];
    FILE *ifile;
    if( cflag ) {
      if( CR_file[0] != '\0' ) {
    
	if(( ifile = fopen(CR_file,"r")) == NULL ) {
	  printf("\n\n Can't open Copyright file %s\n",CR_file);
	  exit(1);
	}
	IFprintf(outfile,"/*\n *\n *\n");
	while( fgets(iline,179,ifile) ) {
	  IFprintf(outfile," *%s",iline);
	}
#if 0
	IFprintf(outfile,"\n *\n * Generated with LDDK Version %s\n */\n\n",PROG_VERSION);
#else
IFprintf(outfile,"\n *\n * \n*/\n\n");
#endif
	fclose(ifile);
      }
      else 
	IFprintf(outfile,"\
/*
* Generated with LDDK Version %s 
*     (c) 1996,1997 Claus Schroeter (clausi@chemie.fu-berlin.de) all rights reserved,
*         lddk is availiable from the Linux Lab Project
*             http://www.llp.fu-berlin.de/
*         or by anonymus FTP
*             ftp://ftp.llp.fu-berlin.de/pub/linux/LINUX-LAB/LDDK
*
*/
",PROG_VERSION);

    }

}

/*==========================================================================*/

dump_file(char *dir,char *in, char *out)
{

  FILE *infile;
  IFILE *outfile;
  FILE *crfile;

  char c,d, *e;
  char buf[4096],buf2[4096],buf3[4096],buf4[4096];
  int i=0,j=0,k=0,l=0,m=0;
  int cflag=0,isfunction=0;

  char inname[120];
  char outname[120];
  char *fresult;
     
  struct hlist_entry *np;


  sprintf(inname,"%s/%s/%s.t",TEMPLATE_PATH,dir,in);
  if( in[strlen(in)-1] == 'c' || in[strlen(in)-1] == 'h' ) {
   sprintf(outname,"%s/%s/%s.%c",DEVEL_PATH,dir,out,in[strlen(in)-1]); cflag=1;
  }
  else {
   sprintf(outname,"%s/%s/%s",DEVEL_PATH,dir,out); cflag=0;
  }

  if(( infile = fopen(inname,"r")) == NULL ){
    printf("\n\n Cant open %s/%s\n Please set your template path correctly\n",TEMPLATE_PATH,in);
    exit(1);
  }

  if(( outfile = IFopen(outname)) == NULL ){
    printf("Cant open output %s\n",out);
    exit(1);
  }

  dump_CRNotice( outfile, cflag );

  while( (c = fgetc(infile)) != EOF ){  

    if( c == '%' ){
      if(( d = fgetc(infile) ) == EOF ) break;
      if( d == '<' ){
        i=0;
        isfunction=0;
        buf3[0]=0;
        while(( d = fgetc(infile)) != EOF && d != '>' ){
	   if( d == '|' ) {
	      k=0;  
	      while(( d = fgetc(infile)) != EOF && d !='>' )
		buf2[k++] = d;
	      break;
	   }
	   if( d == '?' ){
	     k=0;l=0;
	     while(( d = fgetc(infile)) != EOF && d !=':' ) buf3[l++]= d; 
	     while(( d = fgetc(infile)) != EOF && d !='>' ) buf2[k++]= d;
	     break; 
	   }

	   /* function call */
	   if( d == '(' ){
	      buf[i++] = '\0'; /*store function name*/
	      k=0;
	      while(( d = fgetc(infile)) != EOF && d !=')' ) buf3[k++]= d;
	      buf3[k++] = '\0'; 
	      m=0;
	      while(( d = fgetc(infile)) != EOF && d !='>' ) {
		if( m==0 ) 
                    while( (d = fgetc(infile)) == ' ' || d =='\n' || d == '\t' ); 
		buf4[m++] = d;
	      }
	      buf4[m++] = '\0';
	      /* now call function */
	      fresult = do_function( buf, buf3 ,buf4 );
              if(fresult) 
		IFputs(fresult,outfile);

	      isfunction=1;
	      break;
	   }
	   buf[i++] = d;
	}
	if( !isfunction ) {
       	buf[i++] = '\0';
	buf2[k++] = '\0';
	buf3[l++] = '\0'; 
	/*printf("macro '%s' replacement '%s' \n",buf,buf2);*/
	 
	if(( np = hash_lookup(buf) ) == NULL ){
	  /*printf("\n Warning:unexistent macro %s ignored\n",buf);*/
	  if ( buf2[0] != 0 ) IFputs( buf2, outfile ); 
	}
	else
	  {
	    /*printf("macro %s -> %s\n",buf, np->val );*/
	    if( buf3[0] == 0 ) IFputs( np->val, outfile);
	    else IFputs( buf3 ,outfile ); 
	  }
	
        } 
      } else {
	  IFputc(c,outfile);
	  IFputc(d,outfile);
	}
    } else {   
    IFputc(c,outfile);

    }


  }

  fclose(infile);
  IFclose(outfile);

}

/*==========================================================================*/


void add_objects(char *name)
{
  char buf[80];
  sprintf(buf,"%s.o ",name);
  DScat(driver_objects,(DynString *) buf);	
}

void add_sources(char *name)
{
  char buf[80];
  sprintf(buf,"%s.c ",name);
  DScat(driver_sources,(DynString *) buf);	
}

void add_libobjects(char *name)
{
  char buf[80];
  sprintf(buf,"%s.o ",name);
  DScat(lib_objects,(DynString *) buf);	
}

void add_libsources(char *name)
{
  char buf[80];
  sprintf(buf,"%s.c ",name);
  DScat(lib_sources,(DynString *)buf);	
}

void generate_code(char *name)
{

  char buf[1024];
  char buf2[512];
  char buf3[30];
  char *main_name = (char *) hash_get("driver.name");
  int submod=0;

  IFILE *out1,*out2;

  /*name = hash_get("driver.name");*/
  if(driver_minor == 0) {
    submod = 0;
    /*hash_install("driver.name",name);*/
  } else {
    submod = 1;
  }
  hash_install("driver.modname",name);

  printf("\n\n Generating Driver '%s' :",name);
     if(submod) printf("(minor=%d)",driver_minor);
  /* 
   * generating file operation struct
   *
   *
   */
  sprintf(buf2,"%s/include/%s_Proto.h",DEVEL_PATH,name);  
  if(( out1 = IFopen(buf2)) == NULL ){
	printf("Error: can't open %s \n",buf2);
	exit(1);
  }  
  IFprintf(out1,"#ifdef __KERNEL__\n");

  printf("\n\t fops, ");
  buf[0]='\0';
  sprintf(buf,"struct file_operations %s_fops = { \n", name );
  if( driver_methods & LDDK_SEEK ) {
	sprintf(buf2,"%s_seek, \n",name); strcat(buf,buf2);
	IFprintf(out1,"extern int %s_seek (struct inode *, struct file *, off_t, int); \n",name);
  }  else strcat(buf,"NULL, \n");
  if( driver_methods & LDDK_READ ) {
	sprintf(buf2,"%s_read, \n",name); strcat(buf,buf2);
	IFprintf(out1,"extern int %s_read (struct inode *, struct file *, char *, int); \n",name);
  }  else strcat(buf,"NULL, \n");
  if( driver_methods & LDDK_WRITE ) {
	sprintf(buf2,"%s_write, \n",name); strcat(buf,buf2);
	IFprintf(out1,"extern int %s_write (struct inode *, struct file *, char *, int); \n",name);
  }  else strcat(buf,"NULL, \n");
  if( driver_methods & LDDK_READDIR ) {
	sprintf(buf2,"%s_readdir, \n",name); strcat(buf,buf2);
	IFprintf(out1,"extern int %s_readdir (struct inode *, struct file *, struct dirent *, int); \n",name);
  }  else strcat(buf,"NULL, \n");
  if( driver_methods & LDDK_SELECT ) {
	sprintf(buf2,"%s_select, \n",name); strcat(buf,buf2);
	IFprintf(out1,"extern int %s_select ( struct inode *, struct file *, int, select_table *); \n",name);
  }  else strcat(buf,"NULL, \n");
  if( driver_methods & LDDK_IOCTL ) {
	sprintf(buf2,"%s_ioctl, \n",name); strcat(buf,buf2);
	IFprintf(out1,"extern int %s_ioctl ( struct inode *, struct file *, unsigned int, unsigned long ); \n",name);
  }  else strcat(buf,"NULL, \n");
  if( driver_methods & LDDK_MMAP ) {
	sprintf(buf2,"%s_mmap, \n",name); strcat(buf,buf2);
	IFprintf(out1,"extern int %s_mmap (struct inode *, struct file *, struct vm_area_struct *); \n",name);
  }  else strcat(buf,"NULL, \n");

  if( driver_methods & LDDK_OPEN_SUB ) {
	sprintf(buf2,"%s_fops_init, \n",main_name); strcat(buf,buf2);
	IFprintf(out1,"extern int %s_fops_init (struct inode *, struct file *); \n",main_name);
  } 
  else if( driver_methods & LDDK_OPEN ) {
	sprintf(buf2,"%s_open, \n",name); strcat(buf,buf2);
	IFprintf(out1,"extern int %s_open (struct inode *, struct file *); \n",name);
  }  else strcat(buf,"NULL, \n");


  if( driver_methods & LDDK_CLOSE ) {
	sprintf(buf2,"%s_close, \n",name); strcat(buf,buf2);
	IFprintf(out1,"void %s_close (struct inode *, struct file *); \n",name);
	  }  else strcat(buf,"NULL, \n");

  strcat(buf,"};\n");
  hash_install("driver.fops_struct",buf);
  IFprintf(out1,"#endif \n");
  /*fflush(out1);*/
  IFclose(out1);

  /* generating prototypes */
  printf("\n\t prototypes, ");


  sprintf(buf2,"#include <%s_Proto.h>",name);
  hash_install("driver.include-proto",buf2);
  sprintf(buf2,"#include <%s_debug.h>",main_name);
  hash_install("driver.include-debug",buf2);

   if( driver_methods & LDDK_IOCTL ){
    sprintf(buf2,"#include <%s_ioctl.h>",name);
    hash_install("driver.include-ioctl",buf2);
    sprintf(buf2,"#include <%s_types.h>",name);
    hash_install("driver.include-types",buf2);
  }

   
   
  /*
   * generate files
   *
   */

  printf("\n\t methods, ");
  if( driver_methods & LDDK_SEEK ){
	sprintf(buf2,"%s_seek",name);
	dump_file("driver","seek.c",buf2);
	add_objects(buf2);
	add_sources(buf2);
  }

  if( driver_methods & LDDK_READ ){
	sprintf(buf2,"%s_read",name);
	dump_file("driver","read.c",buf2);
	add_objects(buf2);
	add_sources(buf2);
  }
  if( driver_methods & LDDK_WRITE ){
	sprintf(buf2,"%s_write",name);
	dump_file("driver","write.c",buf2);
	add_objects(buf2);
	add_sources(buf2);
  }
  if( driver_methods & LDDK_READDIR ){
	sprintf(buf2,"%s_readdir",name);
	dump_file("driver","readdir.c",buf2);
	add_objects(buf2);
	add_sources(buf2);
  }
  if( driver_methods & LDDK_SELECT ){
	sprintf(buf2,"%s_select",name);
	dump_file("driver","select.c",buf2);
	add_objects(buf2);
	add_sources(buf2);
  }
  if( driver_methods & LDDK_IOCTL ){

        printf("\n\t ioctl functions: ");
        generate_iofuncs(name);

	sprintf(buf2,"%s_ioctl",name);
	dump_file("driver","ioctl.c",buf2);
	add_objects(buf2);
	add_sources(buf2);
  }
  if( driver_methods & LDDK_MMAP ){
	sprintf(buf2,"%s_mmap",name);
	dump_file("driver","mmap.c",buf2);
	add_objects(buf2);
	add_sources(buf2);
  }
  if( driver_methods & LDDK_OPEN ){
	sprintf(buf2,"%s_open",name);
	dump_file("driver","open.c",buf2);
	add_objects(buf2);
	add_sources(buf2);
  }
  if( driver_methods & LDDK_CLOSE ){
	sprintf(buf2,"%s_close",name);
	dump_file("driver","close.c",buf2);
	add_objects(buf2);
	add_sources(buf2);
  }
  printf("\n\t extra interfaces:");
  if( driver_methods & LDDK_PROCINFO ){
        printf("\n\t\t procinfo, ");

        sprintf(buf2,"#define LDDK_USE_PROCINFO 1\n");
        hash_install("driver.procinfo-defs",buf2);
	sprintf(buf2,"%s_procinfo",name);
	dump_file("driver","procinfo.c",buf2);
	add_objects(buf2);
	add_sources(buf2);
  }
  if( driver_methods & LDDK_SYSCTL ){

        printf("\n\t\t sysctl, ");
        generate_sysvars(name);

        sprintf(buf2,"#define LDDK_USE_SYSCTL 1\n#include <linux/sysctl.h>\n\
#ifdef __KERNEL__\n\
#include <Gpib_sysctl.h>\n\
#endif\n");
        hash_install("driver.sysctl-defs",buf2);

	sprintf(buf2,"%s_sysctl",name);
	dump_file("include","sysctl.h",buf2);

	sprintf(buf2,"%s_sysctl",name);
	dump_file("driver","sysctl.c",buf2);
	add_objects(buf2);
	add_sources(buf2);

        
  }
  printf("\n\t core, ");fflush(stdout);

  sprintf(buf,"\
    int %s_major = %s_MAJOR; \n\
  ",main_name,main_name);
  hash_install("driver.global-config",buf);

  sprintf(buf,"\
    if( register_chrdev(%s_major,\"%s\",&%s_fops) ){\n\
      printk(\"can't get Major %%d\\n\",%s_major);\n\
      return(-EIO);\n\
    }\n\
  \n",main_name,main_name,name,
      main_name);
  hash_install("driver.register-major",buf);


  sprintf(buf,"\
    if( unregister_chrdev(%s_major,\"%s\") != 0 ){\n\
      printk(\"can't unregister %s, device busy \\n\");\n\
    } else {\n\
      printk(\"%s successfully removed\\n\");\n\
    }\n\
  \n",main_name,main_name,main_name,main_name);
  hash_install("driver.unregister-major",buf);



  if( driver_methods & LDDK_CORE ){
	sprintf(buf2,"%s_core",name);
	dump_file("driver","core.c",buf2);
	add_objects(buf2);
	add_sources(buf2);
  }




  if( !submod) {
    printf("\n\t headers,");
    hash_install("driver.custom-headers",DSget(driver_headers));

    sprintf(buf2,"%s_debug",main_name);
    dump_file("include","debug.h",buf2); add_objects(buf2);
    dump_file("driver","debug.c",buf2); add_sources(buf2);

    sprintf(buf2,"%s_error",main_name);
    /*dump_file("include","error.h",buf2); */
    dump_file("driver","error.c",buf2); add_objects(buf2); add_sources(buf2);

    sprintf(buf2,"%s_P",main_name);
    dump_file("include","P.h",buf2);
  } else {
    printf("\n\t submodule headers, ");
    sprintf(buf2,"%s_fops",name);
    dump_file("include","fops.h",buf2); 

    sprintf(buf,"\n#include <%s_fops.h>" , name ); DScat(submod_includes,(DynString *) buf);
    sprintf(buf,"\n\t case %d:\n\t\t file->f_op = &%s_fops;\n\t\t break;\n\t\t\t" 
    ,driver_minor, name ); DScat(submod_cases,(DynString *) buf);
    

  }
}


do_on_end(){
   char buf2[80];
   char *main_name = hash_get("driver.name");

   hash_install("submod-includes",DSget(submod_includes));
   hash_install("submod-cases",DSget(submod_cases));

   if( strlen(DSget(submod_includes)) > 1 ){
  printf("\n\t fops_init");
    sprintf(buf2,"%s_fops_init",main_name);
    dump_file("driver","fops_init.c",buf2); 
    add_sources(buf2); add_objects(buf2);

   }
  printf("\n\t Makefiles");
   hash_install("driver.make-objects",DSget(driver_objects));
   hash_install("driver.make-sources",DSget(driver_sources));
   dump_file("driver","Makefile","Makefile");


   hash_install("lib.make-objects",DSget(lib_objects));
   hash_install("lib.make-sources",DSget(lib_sources));
   dump_file("lib","Makefile","Makefile");

   /*hash_install("lang.make-objects",DSget(lib_objects));
   hash_install("lang.make-sources",DSget(lib_sources));*/
   dump_file("language","Makefile","Makefile");


   sprintf(buf2,"%s_lib",main_name);
   hash_install("lib.headers",DSget(lib_headers));
   dump_file("include","lib.h",buf2);

   dump_file("","Makefile","Makefile");

   dump_file("language","test.tcl","test.tcl");

  printf("\n\n");
}



/*==========================================================================*/

sysctl_var_t driver_sysvars[MAX_SYSVARS];
int          sysvar_count = 0;

int store_sysvar(char *aname, int atype, int amode, unsigned long adefault, char *handler, char *strategy ) 
{
  int i;
  extern char *yytext;

  if(sysvar_count < MAX_SYSVARS) {

    for(i=0;i<sysvar_count;i++) {
      if( strcmp(aname,driver_sysvars[i].name) == 0 ) {
	yytext = aname;
	yyerror(" Name conflict: sysctl variable multiple defined ");
	exit(1);
	}
    }




    driver_sysvars[sysvar_count].name = strsave( aname );
    driver_sysvars[sysvar_count].type = atype;
    driver_sysvars[sysvar_count].mode = amode;
    
    driver_sysvars[sysvar_count].handler 
            = (handler != NULL ? strsave( handler ) : NULL);
    
    driver_sysvars[sysvar_count].strategy 
            = (strategy != NULL ? strsave( strategy ) : NULL );

    switch(atype) {
    case SYSCTL_INT:
    case SYSCTL_UINT:
      driver_sysvars[sysvar_count].default_i = (int) adefault;
      break;
    case SYSCTL_STRING:
      driver_sysvars[sysvar_count].default_s = strsave( (char *) adefault );
      driver_sysvars[sysvar_count].default_slen = strlen( driver_sysvars[sysvar_count].default_s );
      printf("string is %s ",driver_sysvars[sysvar_count].default_s);
      break;
    default:
      printf("Internal Error: illegal sysvar type %d \n");
      exit(1);
      break;
    }


    sysvar_count++;
    printf("sysctl var '%s' [%d] stored\n",aname,sysvar_count-1);
  } else {
    printf("Error: sysvar capacity exceeded :-( please change MAX_SYSVARS ");
    exit(1);
  }
}





/*==========================================================================*/

ioctl_func_t driver_iofuncs[MAX_IOFUNCS];
int iofunc_count = 0;

store_iofunct(char *name,char *par,char *code)
{
  if(iofunc_count < MAX_IOFUNCS){
  driver_iofuncs[iofunc_count].name = strsave(name);
  driver_iofuncs[iofunc_count].parlist = strsave(par);
  driver_iofuncs[iofunc_count].code = strsave(code);
  iofunc_count++;
  /*printf("ioctl '%s' (%d) stored\n",name,iofunc_count-1);*/
  } else {
    printf("Error: function capacity exceeded :-( please change MAX_IOFUNCS ");
    exit(1);
  }
}


restore_state()
{
  iofunc_count    = 0;
  sysvar_count    = 0;
  driver_methods  = 0;
  driver_minor    = 0;
  driver_major    = 0;
  driver_minor_to = 0;
}


generate_sysvars(char *driver_name) 
{
  int i,j,k;
  char *main_name = hash_get("driver.name");
  char buf[2*4096];
  char buf2[2*4096];
  char buf3[2*4096];

  

  char dummy[1024];
  char handler[120];
  char strategy[120];


  char *p,*q,*r;


  p = buf;
  q = buf2;
  r = buf3;

  /* store implicit sysvars as debugging etc */

  store_sysvar("dbgMask",SYSCTL_UINT,0644,0,NULL,NULL);   

  /* generate code templates */

  if( driver_methods & LDDK_SYSCTL ) { 
    for(i=0;i<sysvar_count;i++) {

      strcpy( dummy,driver_sysvars[i].name );

      r += sprintf(r, "\n /* ------ Global Definitions for %s */\n\n",driver_sysvars[i].name);
           

      switch( driver_sysvars[i].type ) {
      case SYSCTL_INT:
      case SYSCTL_UINT: 

	if( driver_sysvars[i].handler ) {
	  r += sprintf(r,"extern int %s ( ctl_table *table, int write, struct file *filp, void *buffer, size_t *lenp );\n", 
		       driver_sysvars[i].handler);
	  sprintf(handler,"&%s",driver_sysvars[i].handler);
	} else
	  sprintf(handler,"&%s_dointvec",main_name);

	if( driver_sysvars[i].strategy ) {
	  r +=  sprintf(r,"extern int %s ( ctl_table *table, int *name, int nlen, void *oldval, size_t *oldlenp, void *newval, size_t newlen, void **context );\n",
			driver_sysvars[i].strategy);
	  sprintf(strategy,", &%s",driver_sysvars[i].strategy);
	} else
	  sprintf(strategy,", NULL ");


	printf("\n\t\t\t %s \t (type=int,modes=0%o ),",driver_sysvars[i].name, driver_sysvars[i].mode);
	p += sprintf(p,"{ SYSCTL_%s, \"%s\", &%s, sizeof(int), \n\t\t 0%o, NULL, %s %s },\n ", 
                   
		     uppercase( dummy ),

		     driver_sysvars[i].name, driver_sysvars[i].name, 
		     driver_sysvars[i].mode, handler, strategy
		     ); 
	if( strcmp( driver_sysvars[i].name,"dbgMask") ) {
	  q += sprintf(q,"int %s = %d;\n ",
		       driver_sysvars[i].name, 
		       driver_sysvars[i].default_i 
		       ); }
	r += sprintf(r,"extern %s int %s;\n#define SYSCTL_%s %d\n ",
		     (driver_sysvars[i].type == SYSCTL_UINT ? "unsigned" : "" ),  
		     driver_sysvars[i].name,
		     uppercase( dummy ),
		     i+1   
		     ); 
          
	break;

      case SYSCTL_STRING: 

	if( driver_sysvars[i].handler ) {
	  r += sprintf(r,"extern int %s ( ctl_table *table, int write, struct file *filp, void *buffer, size_t *lenp );\n", 
		       driver_sysvars[i].handler);
	  sprintf(handler,"&%s",driver_sysvars[i].handler);
	} else
	  sprintf(handler,"&%s_dostring",main_name);

	if( driver_sysvars[i].strategy ) {
	  r +=  sprintf(r,"extern int %s ( ctl_table *table, int *name, int nlen, void *oldval, size_t *oldlenp, void *newval, size_t newlen, void **context );\n",
			driver_sysvars[i].strategy);
	  sprintf(strategy,", &%s",driver_sysvars[i].strategy);
	} else
	  sprintf(strategy,", &%s_sysctl_string",main_name);


	printf("\n\t\t\t %s \t (type=string,modes=0%o,len=%d),",driver_sysvars[i].name, driver_sysvars[i].mode,driver_sysvars[i].default_slen );

	p += sprintf(p,"{ SYSCTL_%s, \"%s\", &%s, %d, \n\t\t 0%o, NULL, %s %s },\n ", 
		     uppercase( dummy ),
		     driver_sysvars[i].name, 
                     driver_sysvars[i].name, 
                     driver_sysvars[i].default_slen, 
		     driver_sysvars[i].mode, 
                     handler, 
                     strategy
		     ); 

	q += sprintf(q,"char %s[] = \"%s\";\n ",
		     driver_sysvars[i].name, 
		     driver_sysvars[i].default_s 
		     ); 

	r += sprintf(r,"extern char %s[];\n#define SYSCTL_%s %d\n ",
		     driver_sysvars[i].name,
		     uppercase( dummy ),
		     i+1   
		     ); 

	break;


      default:
	printf("Error: illegal sysvar type ");
	break;
      }

    



    }
    hash_install("driver.sysctl.table"  ,buf);  
    hash_install("driver.sysctl.vars"   ,buf2);  
    hash_install("driver.sysctl.headers",buf3);  
  }

}


generate_iofuncs(char *driver_name)
{

 int i,j,k;
 char cases[10*1024];
 char buf[4*1024],macro[60];
 char *name,*par,*code;
 char *s1,*s2;
 char *argv[20];
 char *main_name = hash_get("driver.name");
 int  argc;


 IFILE *codefile,*header,*types,*libfile;

 sprintf(buf,"%s/include/%s_ioctl.h",DEVEL_PATH,driver_name);
   if(( header = IFopen(buf) ) == NULL ){
     printf("Error can't open %s \n",buf);
     exit(1);
   }
 sprintf(buf,"%s/include/%s_types.h",DEVEL_PATH,driver_name);
   if(( types = IFopen(buf) ) == NULL ){
     printf("Error can't open %s \n",buf);
     exit(1);
   }



   if( driver_methods & LDDK_OPEN ) 
    IFprintf(swigfile,"extern int %s_open(char *name);\n",driver_name);

   if( driver_methods & LDDK_CLOSE ) 
    IFprintf(swigfile,"extern int %s_close(int fd);\n",driver_name);

   if( driver_methods & LDDK_READ ) 
    IFprintf(swigfile,"extern int %s_read(int fd, char *buf, size_t count);\n",
            driver_name);

   if( driver_methods & LDDK_WRITE ) 
    IFprintf(swigfile,"extern int %s_write(int fd, char *buf, size_t count);\n",
            driver_name);

   sprintf(buf,"%s_file",driver_name);
	dump_file("language","file.c",buf);

   
 strcpy(cases,"");

 for(i=0;i<iofunc_count;i++){
   name = driver_iofuncs[i].name;
   par  = driver_iofuncs[i].parlist;
   code = driver_iofuncs[i].code;

   while( *par == ' ' || *par == '\t' ) par++;
   argc = extract_idents(par,argv);

   printf("\n\t\t %s_%s.c ",driver_name,name);
   sprintf(buf,"%s/driver/%s_%s.c",DEVEL_PATH,driver_name,name);
   if(( codefile = IFopen(buf) ) == NULL ){
     printf("Error can't open %s \n",buf);
     exit(1);
   }


   sprintf(buf,"%s_%s",driver_name,name);add_sources(buf); add_objects(buf);   
   IFprintf(codefile,"#include <%s_P.h>\n\n\n",main_name);
   IFprintf(codefile,"%s_%s( ",driver_name,name);
   IFprintf(codefile,"%s",par);
   IFprintf(codefile," )\n{\n%s\n}\n",code);
   IFclose(codefile);


   sprintf(buf,"%s_%s.c ",driver_name,name);
   DScat(lib_sources,(DynString *)buf);
   sprintf(buf,"%s_%s.o ",driver_name,name);
   DScat(lib_objects,(DynString *)buf);


   strcpy(macro,name);
   IFprintf(header,"\n#define %s \t\t %d",uppercase(macro),i);


   sprintf(buf,"%s/lib/%s_%s.c",DEVEL_PATH,driver_name,name);
   if(( libfile = IFopen(buf) ) == NULL ){
     printf("Error can't open %s \n",buf);
     exit(1);
   }



   dump_CRNotice(libfile, 1);


   IFprintf(libfile,"#include <sys/ioctl.h>\n");   
   IFprintf(libfile,"#include <%s_ioctl.h>\n",driver_name);   
   IFprintf(libfile,"#include <%s_types.h>\n\n",driver_name);   

    
   IFprintf(swigfile,"extern int %s_%s(int fd%c%s);\n",
	   driver_name,name,(argc>0?',':' '),par);
    
    
   IFprintf(libfile,"\
        %s_%s(int fd%c%s)\n\
        {\n\
          extern int errno;\n\
          %s_par_t arg;\n",driver_name,name,(argc>0?',':' '),par,name);

   for(k=0;k<argc;k++){
     /*printf("\n extr:%d=%s",k,argv[k]);*/
     IFprintf(libfile,"             arg.%s = %s;\n",argv[k],argv[k]);
   }


   IFprintf(libfile,"\t\tioctl(fd,%s,(void *) &arg );\n\
          errno = arg.error;
          return( arg.retval );          
        }\n\n\n\
        ", macro,name);
   IFclose(libfile);

   

   IFprintf(types,"\ntypedef struct %s_par {\n",name);
   buf[0]='\0';
   for( j=0;j<strlen(par);j++)
     if(par[j] ==',') IFprintf(types,";\n");
     else
       IFprintf(types,"%c",par[j]);
   IFprintf(types,";\n");
   IFprintf(types,"int error;\n");
   IFprintf(types,"unsigned long retval;\n");
   IFprintf(types,"\n} %s_par_t ;\n\n",name);

   sprintf(buf,"      case %s:\n",macro);strcat(cases,buf);
   sprintf(buf,"\
              if( verify_area(VERIFY_READ, (void *) arg, sizeof(%s_par_t))){\n\
                 DBGout(); return(retval); \n\
              }\n\
              if( verify_area(VERIFY_WRITE, (void *) arg, sizeof(%s_par_t))){\n\
                 DBGout(); return(retval); \n\
              }\n\
              argp = (void *) kmalloc( sizeof(%s_par_t) +1 ,GFP_KERNEL );\n\
              memcpy_fromfs( (void *) argp, (%s_par_t *) arg, sizeof(%s_par_t));\n", 
	      name,
	      name,
	      name,
	      name,name );strcat(cases,buf);
   sprintf(buf,"\
              ((%s_par_t *) argp)->retval = %s_%s( ",name,driver_name,name); 
   strcat(cases,buf);

   for(k=0;k<argc;k++){
      sprintf(buf," ((%s_par_t *) argp)->%s",name, argv[k] );strcat(cases,buf);
      if( k+1 < argc ){
	sprintf(buf,", \n\t\t\t\t");strcat(cases,buf);
      }
   }

   sprintf(buf," );\n");strcat(cases,buf); 
   sprintf(buf,"\
              ((%s_par_t *) argp)->error = %s_errno;\n",name,main_name);
   strcat(cases,buf); 
   

   sprintf(buf,"\
              memcpy_tofs( (%s_par_t *) arg, (void *) argp, sizeof(%s_par_t));\n\
              kfree(argp);
              break;\n",
	   name,name);
   strcat(cases,buf);
 }

 hash_install("driver.ioctl-case-functions",cases);
 hash_install("driver.lib-sources",DSget(lib_sources));
 hash_install("driver.lib-objects",DSget(lib_objects));

  

 IFclose(header);
 IFclose(types);

}


/*==========================================================================*/

int extract_idents(char *par,char **argv)
{
  int i,j,k;
  char c;
  char *buf;
  char buf2[1024];
  int  pargc=0;
  
  buf = par;
  
  i=0;
  while( buf[i] != '\0' && i < strlen(buf) ){
    while( (c=buf[i]) == ' ' || c== '\t' || c == ',' ) i++;
    if( buf[i] == '\0' ) return 0;
    j=0;
    while( (c=buf[i]) != ' ' && c != '\t' && c !='\0' && c != ',' ){
      buf2[j++] = c; i++;
    }
    buf2[j++] = '\0';
    while( (c=buf[i]) == ' ' || c== '\t' ) i++; /* jc */
    if ( buf[i] == ',' || buf[i] == '\0' ){
      for(k=0;k<strlen(buf2);)
	if( buf2[k] == '*' ) k++;
        else 
	  break;
      argv[pargc] = strsave( &buf2[k] );
      /*printf("arg[%d] =%s \n",pargc,argv[pargc]);*/
      pargc++;
    }
  }
  return pargc;

}


/*==========================================================================*/

char *uppercase(char *s)
{
  int i;
  for(i=0;i<strlen(s);i++)
    s[i] = toupper(s[i]);
  
  return s;
}

/*==========================================================================*/

make_sourcetree(){

char *driver_name = hash_get("driver.name");
char cmd[512];
char buf[512];

strcpy(devel_path,driver_name);

printf("\n\nGenerating source-tree in ./%s\n",devel_path);fflush(stdout);
sprintf(cmd,"mkdir %s",devel_path);
system(cmd);
sprintf(cmd,"mkdir %s/driver",devel_path);
system(cmd);
sprintf(cmd,"mkdir %s/include",devel_path);
system(cmd);
sprintf(cmd,"mkdir %s/lib",devel_path);
system(cmd);
sprintf(cmd,"mkdir %s/language",devel_path);
system(cmd);
sprintf(cmd,"mkdir %s/doc",devel_path);
system(cmd);


/* open file for SWIG wrapper generator */
   
   sprintf(buf,"%s/language/%s.i",DEVEL_PATH,driver_name);
   if(( swigfile = IFopen(buf) ) == NULL ){
     printf("Error can't open %s \n",buf);
     exit(1);
   }
   IFprintf(swigfile,"%%module %s\n",driver_name);
   IFprintf(swigfile,"%%{ \n");
   /*IFprintf(swigfile,"#include <sys/ioctl.h>\n");*/   
   IFprintf(swigfile,"#include <%s_P.h>\n",driver_name);   
   /*IFprintf(swigfile,"#include <%s_ioctl.h>\n",driver_name); */  
   /* IFprintf(swigfile,"#include <%s_types.h>",driver_name); */
   IFprintf(swigfile,"\n%%}\n\n");
   
   IFprintf(swigfile,"//%%include tclsh.i        // include init code for tclsh rebuild \n");



}



/*==========================================================================*/



init_states(){

  strcpy(devel_path,"./tmp");
  strcpy(template_path,DEFAULT_TEMPLATE_PATH);
  strcpy(include_path,DEFAULT_INCLUDE_PATH);

  driver_objects= DSinit();
  driver_sources= DSinit();
  driver_headers= DSinit();
  lib_objects=DSinit();
  lib_sources=DSinit();
  lib_headers=DSinit();

  submod_includes = DSinit();
  submod_cases = DSinit();

  doc_buffer = DSinit();
}

release_states() {
  DSfree(driver_objects);
  DSfree(driver_sources);
  DSfree(driver_headers);
  DSfree(lib_objects);
  DSfree(lib_sources);
  DSfree(lib_headers);
  DSfree(submod_includes);
  DSfree(submod_cases);
  if( swigfile != NULL ) IFclose(swigfile);
}


init_main(){
  char *date;
  time_t t;

  time(&t);
  date = ctime(&t);
  hash_install("date",date);
  hash_install("version",PROG_VERSION);


}







