%{
#include "hash.h"
#include "y.tab.h"
#include "dstring.h"

int yylineno=1;int istate=0;

int noline = 0;

DynString *codebuf = NULL;

char tbuf[120];
char *fpath;
char actual_filename[120] = "base";

/* this is for the include statement */


#define MAX_INCLUDE_DEPTH 25

YY_BUFFER_STATE include_stack[MAX_INCLUDE_DEPTH];
int lineno_stack[MAX_INCLUDE_DEPTH];
char *pfile_stack[MAX_INCLUDE_DEPTH];
int include_stack_ptr = 0;

extern char *pfile;
extern char include_path[];

extern char *find_file(char *,char *);
extern DynString *doc_buffer;
int doc_stat;
%}


%x incl
%x COMMENT
%x DOC
%x DOCSEE
%x CODE
%x STRING
%x PARAMETER

IDENT [A-Za-z_][A-Za-z_0-9]*

%%

"/*"             BEGIN(COMMENT);
<COMMENT>"\n"    { yylineno++; }
<COMMENT>[^*]
<COMMENT>"*"
<COMMENT>"*"+"/" BEGIN(INITIAL);


"/@"		{BEGIN(DOC); DScpy(doc_buffer,""); 
                 doc_stat = 1;}
<DOC>^[ \t]*
<DOC>[^@]*	{macro_strcat(doc_buffer,yytext);}
<DOC>"@"+[^@/]* {macro_strcat(doc_buffer,yytext);}
<DOC>"@"+"/"	BEGIN(INITIAL);


include               BEGIN(incl);
<incl>[ \t]*      /* eat the whitespace */
<incl>[^ \t\n]+   { /* got the include file name */
                   if ( include_stack_ptr >= MAX_INCLUDE_DEPTH )
                       {
                       fprintf( stderr, "Includes nested too deeply" );
                       exit( 1 );
                       }

                   include_stack[include_stack_ptr] =
                       YY_CURRENT_BUFFER;
		   lineno_stack[include_stack_ptr] = yylineno;
		   pfile_stack[include_stack_ptr] = pfile;
                   pfile = strsave(yytext);
		   include_stack_ptr++;	
		   yylineno=1;

		   if(( fpath = find_file(yytext, include_path)) == NULL){
                       fprintf(stderr,"Can't find file: %s from %s\n",yytext,include_path);
		        exit(1);
		        }
                       printf("Loading %s ...\n",fpath);

		       if(( yyin = fopen( fpath, "r" )) == NULL ){
                       fprintf(stderr,"Can't open include file: %s\n",yytext);
		        exit(1);
                       }
                   
                   yy_switch_to_buffer(
                       yy_create_buffer( yyin, YY_BUF_SIZE ) );

                   BEGIN(INITIAL);
		   yylineno++;
                   }
<<EOF>> {
                   if ( --include_stack_ptr < 0 )
                       {
                       yyterminate();
                       }

                   else
                       {
                       yy_delete_buffer( YY_CURRENT_BUFFER );
		       yylineno = lineno_stack[include_stack_ptr];
		       pfile = pfile_stack[include_stack_ptr];
                       yy_switch_to_buffer(
                            include_stack[include_stack_ptr] );
                       }
         }






module                { return T_MODULE; }
driver                { return T_DRIVER; }
library               { return T_LIBRARY; }
methods               { return T_METHODS; }

open                  { return T_OPEN; }
open-submodules       { return T_OPEN_SUB;}
close                 { return T_CLOSE;}
read                  { return T_READ; }
write                 { return T_WRITE; }
mmap                  { return T_MMAP;}
select                { return T_SELECT; }
ioctl                 { return T_IOCTL; }

procinfo              { return T_PROCINFO; }
sysctl                { return T_SYSCTL; }
int                   { return T_SINT; }
string                { return T_SSTRING; }
access                { return T_ACCESS; }
handler               { return T_HANDLER;}
strategy              { return T_STRATEGY;}

core                  { return T_CORE; }
init                  { return T_INIT;}
cleanup               { return T_CLEANUP;}
func     	      { return T_FUNCTION; }

name                  { return T_NAME; }
major                 { return T_MAJOR; }
minor                 { return T_MINOR; }
type                  { return T_TYPE; }
doc                   { return T_DOC;}
dump                  { return T_DUMP;}

user                  { return T_USER;}
file                  { return T_FILE;}
header                { return T_HEADER; }
library               { return T_LIBRARY; }
returns               { return T_RETURNS; }

custom-objects        { return T_COBJECTS; }
custom-sources        { return T_CSOURCES; }
custom-defines        { return T_CDEFINES; }

configurable          { return T_CONFIG; }
macro                 { return T_MACRO; }
proc-template         { return T_TEMPLATE;}


dump                  { return T_DUMP; }

0[0-9]+          { sscanf(yytext,"0%o",&yylval.ival); return T_NUMBER; }
[0-9]+           { sscanf(yytext,"%d",&yylval.ival);   return T_NUMBER; }
0x[0-9a-fA-F]+   { sscanf(yytext,"0x%x",&yylval.ival); return T_NUMBER; }




"/{"             { BEGIN(CODE); codebuf = DSinit();
		   if( ! noline )
			DSprintf(codebuf,"#line %d \"%s\"",yylineno,pfile); 
		 }
<CODE>[^}]*      { macro_strcat(codebuf,yytext); }
<CODE>"}"+[^}/]* { macro_strcat(codebuf,yytext); }
<CODE>"}"+"/"    {
	           BEGIN(INITIAL); 
                   /*printf("codebuf=0x%lx \n'%s' \n",DSget(codebuf),DSget(codebuf));fflush(stdout);*/
		   yylval.sval = strsave(DSget(codebuf));
                   DSfree(codebuf);
                   codebuf = NULL;
		   return T_CODE;	
		 }


\"               { BEGIN(STRING); }
<STRING>[^"]*    { yylval.sval = strsave(yytext); 
		   strcat(yylval.sval,"");
		   return T_STRING; }
<STRING>\"       { BEGIN(INITIAL);}

\(		 { BEGIN(PARAMETER); istate=0; }
<PARAMETER>[^)]* { yylval.sval = strsave(yytext); 
		   strcat(yylval.sval,"");
		   istate=1;
		   return T_PARAMETER; }
<PARAMETER>\)    { BEGIN(INITIAL);
		   if(istate == 0) { 
			yylval.sval = strsave(" "); 
			return T_PARAMETER;
		   }
		 }

[a-zA-Z][a-zA-Z_0-9]* { yylval.sval = strsave(yytext); 
			strcat(yylval.sval,"");
			return T_IDENT; }


[ \t]+                      
[ \n ]                        { yylineno++; }

.	         { return(yytext[0]); }

%%

/* do a strcat with macro expansion (for macros in code)*/

macro_strcat(DynString *t, char *s)
{
  int i=0;
  int j=0;
  int isfunction=0;
  int k;
  int l=0,m=0;
  char c,d,buf[4096],buf2[4096],buf3[4096],buf4[4096];
  char *fresult;
  extern int yylineno;

  struct hlist_entry *np;

  /*printf("***s='%s'",s);fflush(stdout);*/

  while(( c = s[i++] ) != '\0' ){
	if( c == '%' ){
	  if(( d = s[i++] ) == '\0' ) break;
          if( d == '<' ){
             buf4[0]='\0';
             isfunction=0;
             k=0;
             while(( d = s[i++] ) != '\0' && d != '>' ) {
               /* single alternative */
               if( d == '|' ) {
	         l=0;
		 while (( d = s[i++] ) != '\0' && d != '>' ) 
		   buf3[l++] = d;
		 break;
	       } 
	       /* true or false */
	       if( d == '?' ){
		  l=0;m=0;
		  while(( d = s[i++]) != '\0' && d !=':' ) buf4[m++]= d;
                  while(( d = s[i++]) != '\0' && d !='>' ) buf3[l++]= d;
	          break;
	       }

               /* function call */
	       if( d == '(' ){
	          buf2[k++] = '\0'; /*store function name*/
                  l=0;m=0;
		  while(( d = s[i++]) != '\0' && d !=')' ) buf3[l++]= d;
		  buf3[l++] = '\0';
                  while(( d = s[i]) == ' ' || d =='\n' || d == '\t' ) i++;
		  while(( d = s[i++]) != '\0' && d !='>' ) buf4[m++]= d;
                  buf4[m++] = '\0';
		  /* now call function */
		  fresult = do_function( buf2, buf3, buf4 );
                  if( fresult )
		   for(m=0;m<strlen(fresult);m++) buf[j++]=fresult[m];
		  isfunction=1;
#if 0
	     printf("\n buf2=%s\n buf3=%s\n buf4=%s",buf2,buf3,buf4);
             fflush(stdout);
#endif
                  break;
               } else {
	         buf2[k++] =d;
               }
	     } 

 
             if(!isfunction){
	      buf2[k++] = '\0';
	      buf3[l++] = '\0';
	      buf4[m++] = '\0';
#if 0
	     printf("\n buf2=%s\n buf3=%s\n buf4=%s",buf2,buf3,buf4);
             fflush(stdout);
#endif
	      if(( np = hash_lookup(buf2)) != NULL ){
	         if( buf4[0] == 0 ){
		    for(k=0;k<strlen(np->val);k++) buf[j++]=np->val[k];
		 } else {
		    for(k=0;k<strlen(buf4);k++) buf[j++]=buf4[k];
		 }
	      } else {
	        if ( buf3[0] != '\0')
	              for(k=0;k<strlen(buf3);k++) buf[j++]=buf3[k];		
	      }
             }
	  } else {
	    buf[j++] = c;
            if( d == '\n' ) yylineno++;
            buf[j++] = d;
          }
next:	     
	} else {
	  if( c == '\n' ) yylineno++;
	  buf[j++] = c;
	}
  }
  buf[j++] = '\0';
#if 0
 printf("\n buf=%s",buf);
#endif
 fflush(stdout);


 DScat(t,buf);


}

