%{
#include "common.h"
#include "hash.h"


extern int yylineno;
extern void yyerror(char *);

char gbuf[150];
extern DynString *driver_headers;
extern DynString *lib_headers;
int errstat=0;

int auto_minor = 0;
int doc_enable = 0;
int doc_type=0;
char *pfile ;

int use_sysctl=0;
int use_procinfo=0;

char *sysctl_handler = NULL;
char *sysctl_strategy = NULL;
int  sysctl_access = 0644;

extern FILE *swigfile;

#define RESTORE_SYSCTL  sysctl_handler =  NULL;sysctl_strategy = NULL;sysctl_access =   0644

%}

%union
{ 
 int   ival; 
 char *sval; 
}

%token T_MODULE T_FUNCTIONS T_METHODS T_NAME T_DRIVER 
%token T_OPEN T_OPEN_SUB T_CLOSE T_READ T_WRITE T_CORE T_INIT T_CLEANUP T_MAJOR
%token T_MODULE T_FUNCTION T_IOCTL T_EXIT T_SELECT T_LIBRARY
%token T_USER T_COBJECTS T_CSOURCES T_CDEFINES T_FILE T_HEADER
%token T_MINOR T_CONFIG T_MACRO T_TEMPLATE T_DOC T_LIBRARY T_DUMP T_TYPE
%token T_PROCINFO T_SYSCTL T_SINT T_ACCESS T_HANDLER T_STRATEGY T_SSTRING T_RETURNS
%token T_MMAP
%token T_DUMP

%token T_NUMBER T_STRING T_CODE T_IDENT T_PARAMETER

%type <ival> T_NUMBER
%type <sval> T_STRING
%type <sval> T_CODE
%type <sval> T_IDENT
%type <sval> T_PARAMETER



%%

mainsec:
	modlist deflist { do_on_end (); }
	| error {yyerror("global structure invalid"); }

modlist: 
	T_MODULE T_STRING ':' { hash_install("driver.name",$2); make_sourcetree(); /*dump_doc(4,NULL,NULL);*/ }
	| error {yyerror("illegal 'module' statement"); exit(1);}

deflist:
	section
	| section deflist
	| error { yyerror("deflist:"); } 	

section:
	secelement
	| secelement section
	| error { yyerror("codeclass or section expected"); }
  
secelement: 
	T_DRIVER T_STRING { hash_install("driver.modname",$2);dump_doc(4,NULL,NULL); } 
           driver_options '{' drvmethods '}' {  generate_code($2); 
                                                            restore_state(); }
	| T_MACRO T_STRING T_CODE { hash_install($2,$3); }
	| T_DUMP T_STRING T_STRING T_STRING { dump_file($2,$3,$4); }
	| T_LIBRARY '{' libmethods '}'
	| error { yyerror("illegal codeclass"); }	


driver_options: 
	| '[' driver_option ']'
	| error { yyerror("driver-options:"); }

driver_option:
	doption
	| doption ',' driver_option
	| error { yyerror("doption list empty"); }


doption:
	T_MAJOR '=' T_NUMBER { sprintf(gbuf,"%d",$3);
				hash_install("driver.major", gbuf );
				hash_install("driver.enable-register","#define LDDK_USE_REGISTER 1\n");
				driver_major=$3; }
	| T_MINOR '=' T_NUMBER { sprintf(gbuf,"%d",$3);
				 hash_install("driver.minor", gbuf ); 
				 driver_minor=$3; }
	| T_MINOR              { auto_minor++; driver_minor=auto_minor;
				 sprintf(gbuf,"%d",driver_minor);
				 hash_install("driver.minor", gbuf ); 
				}
	| error { yyerror("not a option"); }	

drvmethods:
	drvmethod
	| drvmethod drvmethods
	| error { yyerror("not a method "); }

drvmethod:
	| T_MACRO T_STRING T_CODE { hash_install($2,$3); }
	| T_CSOURCES T_CODE { hash_install("driver.custom-sources",$2); }
	| T_COBJECTS T_CODE { hash_install("driver.custom-objects",$2); }
	| T_CDEFINES T_CODE { hash_install("driver.custom-defines",$2); }
	| T_FILE T_STRING T_CODE { char *mod;
                                   hash_install("custom.code",$3);
                                   mod = hash_get("driver.modname");
                                   sprintf(gbuf,"%s_%s",mod,$2);
				   dump_file("driver","code.c",gbuf); 
				   add_sources(gbuf); add_objects(gbuf);}
	| T_HEADER T_STRING T_CODE {
				   char *mod;
                                   char gbuf2[60];
   		                   hash_install("header.code",$3);
                                   mod=hash_get("driver.modname");
				   sprintf(gbuf2,"%s_%s",mod,$2);
		                   dump_file("include","header.h",gbuf2);
		                   sprintf(gbuf,"\n#include <%s.h>",gbuf2);
		                   DScat(driver_headers,(DynString *) gbuf);
	}
        | T_OPEN_SUB          { driver_methods |= LDDK_OPEN_SUB; }
	| T_OPEN T_CODE       { driver_methods |= LDDK_OPEN;
			      hash_install("driver.open_code", $2 );}
	| T_READ T_CODE       { driver_methods |= LDDK_READ;
			      hash_install("driver.read_code", $2 );
			      dump_doc(3,"read","int fd,char *buf,int len");}
	| T_WRITE T_CODE       { driver_methods |= LDDK_WRITE;
			      hash_install("driver.write_code", $2 );
			      dump_doc(3,"write","int fd,char *buf,int len");}
        | T_MMAP T_CODE       { driver_methods |= LDDK_MMAP;
				hash_install("driver.mmap_code",$2);}
	| T_SELECT T_CODE     { driver_methods |= LDDK_SELECT;
				hash_install("driver.select_code",$2);
				dump_doc(3,"select","int  n,fd_set *readfds,fd_set *writefds,fd_set *exceptfds,struct timeval *timeout ");
	                      }
	| T_CLOSE T_CODE    { driver_methods |= LDDK_CLOSE;
			      hash_install("driver.close_code", $2 );}
	| T_PROCINFO T_CODE  { if(use_procinfo){
                                  yyerror("Only one procinfo can be used per module!"); 
                                  exit(1);
                               }
                               driver_methods |= LDDK_PROCINFO; use_procinfo++; hash_install("driver.procinfo_code", $2 );}
	| T_CORE   '{' core_methods   '}' { driver_methods |= LDDK_CORE;}
	| T_IOCTL  '{' ioctl_methods  '}' { driver_methods |= LDDK_IOCTL;}
	| T_SYSCTL { /*if(use_sysctl) {
                          yyerror("Only one sysctl can be used per module!"); 
                          exit(1);
                     }*/ } 
                  '{' sysctl_methods '}' { driver_methods |= LDDK_SYSCTL; use_sysctl++; }
	| T_MACRO T_STRING T_CODE { hash_install($2,$3); }
	| T_FUNCTION dretval T_IDENT T_PARAMETER T_CODE {
	                hash_install("driver.funcname",$3);
			hash_install("driver.func-param",$4);
			hash_install("custom.code",$5);
			dump_file("driver","func.c",$3);
			add_sources($3); add_objects($3);
			dump_doc(1,$3,$4);
          }
	|  error { yyerror("driver method expected"); }

dretval: /* empty */ { hash_install("driver.func-retval","int"); }
	| T_RETURNS T_IDENT { hash_install("driver.func-retval",$2);}
	| T_RETURNS T_IDENT '*' { sprintf(gbuf,"%s *",$2); hash_install("driver.func-retval",gbuf);}
	| error { yyerror("illegal retval statement "); }


libmethods:
	libmethod
	| libmethod libmethods
	| error { yyerror("'lib' methods expected");}



libmethod:
	| T_MACRO T_STRING T_CODE { hash_install($2,$3); }
	| T_FILE T_STRING T_CODE { hash_install("custom.code",$3);
				   dump_file("lib","code.c",$2); 
				   sprintf(gbuf,"%s.c",$2);
				   add_libsources(gbuf); 
				   sprintf(gbuf,"%s.o",$2);
				   add_libobjects(gbuf);}
	| T_HEADER T_STRING T_CODE { hash_install("header.code",$3);
				     dump_file("include","header.h",$2);
				     sprintf(gbuf,"\n#include <%s.h>",$2);
		                     DScat(lib_headers,(DynString *) gbuf); 
                                   }
	| T_FUNCTION retval T_IDENT T_PARAMETER T_CODE {
	                hash_install("lib.funcname",$3);
			hash_install("lib.func-param",$4);
			hash_install("custom.code",$5);
			dump_file("lib","func.c",$3);
			add_libsources($3); add_libobjects($3);
			fprintf(swigfile,"extern %s %s_%s (%s);\n",
				hash_get("lib.func-retval"),hash_get("driver.name"),$3,$4 );
			dump_doc(2,$3,$4);
          }
	| error { yyerror("libmethod expected"); }

retval: /* empty */ { hash_install("lib.func-retval","int"); }
	| T_RETURNS T_IDENT { hash_install("lib.func-retval",$2);}
	| T_RETURNS T_IDENT '*' { sprintf(gbuf,"%s *",$2); hash_install("lib.func-retval",gbuf);}
	| error { yyerror("illegal retval statement "); }

core_methods:
	| core_method
	| core_method core_methods
	| error { yyerror("'core' needs methods"); }

core_method:
	| T_INIT T_CODE {hash_install("driver.init_code", $2 );}
	| T_CLEANUP T_CODE {hash_install("driver.cleanup_code", $2 );}
	|  error { yyerror("not a method of 'core'"); }		



sysctl_methods:
	| sysctl_method
	| sysctl_method sysctl_methods
	| error { yyerror("'sysctl' needs methods"); }

sysctl_method:
	| T_SINT T_IDENT sysctl_options '=' T_NUMBER {
                store_sysvar( $2, SYSCTL_INT , sysctl_access , (unsigned long) $5,sysctl_handler,sysctl_strategy ); 
                RESTORE_SYSCTL;
          }
        | T_SSTRING T_IDENT sysctl_options '=' T_STRING {
                store_sysvar( $2, SYSCTL_STRING , sysctl_access , (unsigned long) $5,sysctl_handler,sysctl_strategy ); 
                RESTORE_SYSCTL;
	}
        | T_HANDLER T_IDENT T_CODE { char *mod;
                                     hash_install("sysctl.handler.code",$3);
                                     hash_install("sysctl.handler.name",$2);
                                     mod = hash_get("driver.name");
                                     sprintf(gbuf,"%s_%s",mod,$2);
				     dump_file("driver","sysctl_handler.c",gbuf); 
				     add_sources(gbuf); add_objects(gbuf);
                                   }
        | T_STRATEGY T_IDENT T_CODE { 
                                     char *mod;
                                     hash_install("sysctl.strategy.code",$3);
                                     hash_install("sysctl.strategy.name",$2);
                                     mod = hash_get("driver.name");
                                     sprintf(gbuf,"%s_%s",mod,$2);
				     dump_file("driver","sysctl_strategy.c",gbuf); 
				     add_sources(gbuf); add_objects(gbuf);
                                    }
        | error { yyerror("not a method of 'sysctl'"); }	

sysctl_options: 
	| '[' sysctl_option ']'
	| error { yyerror("sysctl-options:"); }

sysctl_option:
	soption
	| soption ',' sysctl_option
	| error { yyerror("sysctl option list empty"); }

soption:
	T_ACCESS '=' T_NUMBER    { sysctl_access = $3; }
        | T_HANDLER '=' T_IDENT  { sysctl_handler = $3; }
        | T_STRATEGY '=' T_IDENT { sysctl_strategy = $3; }
	| error { yyerror("not a option"); }	


ioctl_methods:
	| ioctl_method
	| ioctl_method ioctl_methods
	| error { yyerror("'ioctl' need methods"); }

ioctl_method:
	| T_INIT T_CODE {hash_install("driver.ioctl_init",$2);}
	| T_CLEANUP T_CODE {hash_install("driver.ioctl_exit",$2);}
	| T_USER T_CODE {hash_install("driver.ioctl-case-user",$2);}
	| T_FUNCTION T_IDENT T_PARAMETER T_CODE { 
			store_iofunct($2,$3,$4); 
			dump_doc(3,$2,$3);}
	|  error { yyerror("not a method of 'ioctl'"); }		

%%
void yyerror(char *s){

	extern char *yytext;
	extern int yylineno;

	/*printf ("\n*** Error(%s): %s in line %d\n---> %s \n\n",pfile,s,yylineno,yytext);*/
	printf("\n%s:%d: %s -> '%s' \n",pfile,yylineno,s,yytext);
	errstat++;
        if( errstat > 10 ){
		printf("\n\n ** too many errors, exiting...\n");
                exit(1);
        }
}

