/@
   The Maptest module shows how a piece of memory in the kernel space 
   can be mapped to the user memory.
   Xfree uses this method to access the video board memory directly
   from user level via /dev/kmem. 
   It is also possible to map a DMA buffer or whatever.
   'normal' memory is usually not used for mapping so it has to be marked
   as 'reserved' before it can be used for mapping.

   The driver allocates one page of kernel memory that will be 
   initialized with a string. In user space this memory area 
   can be used for reading and writing (see maptest.c)
   maptest supports reading and writing the buffer just by adding 
   a commandline argument:

       maptest "test content"
   
   will show the old content of the buffer and change it to the new value.

   The 'procinfo' method is used to make this buffer visible in the
   /proc/Maptest file.
   
      cat /proc/Maptest

   shows   
      Proc entry from Maptest:
      Message: 'Test Content'
   
   

@/

module "Mapping" : 
driver "Mapping" [ major=34 ] {

        file "globals" /{

           /* this is the global adress for the 'DMA'-buffer */

           unsigned long global_message;

        }/

	core {  
	     init /{ 

                int i;
                 
		extern char *global_message;
		printk("Maptest initialize: ");

                /*  Get one page as global 'buffer' 
                 *  This could be used for DMA or whatever
                 */
                global_message =__get_free_page(GFP_KERNEL);
		if( !global_message ){
		  printk("failed \n");
		  return 0;
		}
                /* mark it as reserved to get mmap() to work */
                for (i = MAP_NR (global_message); 
                      i <= MAP_NR (global_message+PAGE_SIZE-1); i++)
		    {
      			mem_map_reserve (i);
    		}
                /* now fill the buffer with garbage */
		memcpy((void *) global_message,
                      (void *) "This Is a global Message", 25 );
		printk("OK \n");
			
             }/

             cleanup /{
		int i;
                extern char *global_message;

                /* unmark memory */
                for (i = MAP_NR (global_message);  
                   i <= MAP_NR (global_message+PAGE_SIZE-1); i++)
		{
                   mem_map_unreserve (i);
    		}

                if( global_message ){
		   free_page(global_message);
                }
	        printk("Have a nice day! Good bye! \n");
			
             }/
	}

        /*
         * This method will be called whenever the driver is opened
         */
	open /{
		printk("%<driver.name> module has been opened\n");
	}/

        /*
         * This method will be called whenever the driver is closed
         */
	close /{
                extern char* global_message;

		printk("Stupid module has been closed '%s' \n",global_message );
	}/



        mmap /{

           #include <linux/mm.h>
	   #include <asm/io.h>
	   
           struct vm_operations_struct *ops;
           extern unsigned long global_message;
	   unsigned long off;
	   
           printk("\n VMA->flags=0x%lx start=0x%x end=0x%lx len=%d offset=0x%lx \n",
	            vma->vm_flags,vma->vm_start,
		    vma->vm_end,
		    vma->vm_end - vma->vm_start,
		    vma->vm_offset);

	   printk("\n Local Page: 0x%lx (0x%lx) ",global_message,off);	    

             if ( global_message & ~PAGE_MASK ){
	     printk("value not on page boundary");
	     return -ENXIO;
	     }

           printk("message: '%s' \n",global_message);

           if ( remap_page_range(vma->vm_start, 
	        global_message , 
		vma->vm_end - vma->vm_start, 
		vma->vm_page_prot))
           return -EAGAIN;
		
           vma->vm_inode = inode;
	   inode->i_count++;
	   /*vma->vm_ops = ops;*/
	   return 0;
        }/


	procinfo /{
		extern char * global_message;
                char *          p;
	
	        p = buf;
		p += sprintf(p,"Proc entry from %<driver.name>:\n");
		p += sprintf(p,"Message: '%s'\n",global_message);
		retval = p - buf;		
	}/


}


