/* 
 * This example shows how the Ressource Manager can be used for
 * flexible runtime configuration.
 *
 * Of course this module does nothing useful except giving out 
 * some messages :).
 *
 */
/@
  This is a simple demonstration module for the registry mechanism,
  does nothing useful except giving out messages if one of the 
  relevant ressources is modified.
  Before this example can be used, the ressource manager module
  has to be loaded with insmod. Then run rmanip once to install
  the ressources from /etc/kres.conf into the kernel ressource
  database. Load this module into the kernel with insmod and
  run rmanip again.
 
  now play around with the ressources :
 
  /sbin/rmanip --query Registry::flags 
  /sbin/rmanip --set   Registry::flags 55
  /sbin/rmanip --query Registry::flags 
  
  now look through your syslog output with 
  dmesg and you'll see the ressources in action ;-)
@/

module "Registry" : /* this directive sets the global module name */ 

/* To enable the registry feature the 'rmgr' ddl library has to be included
 * here. The rmgr library will install two default ressources into the 
 * Kernel ressource database automatically:
 *
 *  Registry::major type=int     is the major number for the driver
 *  Registry::debug type=int     is the debug level for the module
 *  
 * The values for this ressources can be changed with the rmanip program 
 * that comes with the rmgr package of LDDK. 
 * Realize that the rmgr module has to be loaded to use this module
 *
 */

include rmgr.ddl


driver "Registry" [ major=34 ] {

	/* This statement generates one additional file that defines
         * two global variables and a Callback handler.  
         * The %<...> statements are macros that will be replaced 
         * while the code is generated.
         */

        file "ressources" /{

            void func_a(){
		printk("This is func A\n");
	    }

            void func_b(){
		printk("This is func B\n");
	    }


            int %<driver.name>_base = 0x220 ;
            int %<driver.name>_flags= 0;
            void (* %<driver.name>_func)() = func_a;



            /*
             * The Callback routine can be installed with the ressource
             * to notify a state change or whatever when the ressource
             * is accessed.
             */
            void Flags_ChangeCB(int event,RCS_entry *np){

	       DBGin("Flags_ChangeCB");
	       DBGprint(DBG_DATA,("event=%d",event));
	       switch (event) {
		case RCS_CHANGE_EVENT:
		 printk("%<driver.name>: '%s' changed \n",np->name);
		break;
               }
	       DBGout();
            }	    
        }/

	/* The core init method now installs this variables in the
         * ressource database.  
         *
         */

	core {  
		init /{ 
                        RCS_entry v,*np;
                        extern int %<driver.name>_base,%<driver.name>_flags;
			extern void Flags_ChangeCB();
			extern void (* %<driver.name>_func)();
			extern void func_a(),func_b();

                        /* Look up the ressource in the database 
                         * and set the value if the ressource exists. 
                         *
                         */

                         np = RCS_lookup("%<driver.name>::base");
                         if( np == NULL ){
                            printk("%<driver.name>: 'base' not found, using default\n");
                         }        
                         else 
                         {
                         %<driver.name>_base = np->uval.ival;
                         printk("%<driver.name>: Base = %d\n",
                                                 %<driver.name>_base);
                         }
							
                        /* This will install a linked ressource that 
                         * can be changed at runtime. 
                         *
                         */

                         v.type = RCS_IPTR;
                         v.uval.iptr = &%<driver.name>_flags;
                         np = RCS_install("%<driver.name>::flags",&v);

                         /* now we got the handle to install 
                            a callback routine */

			 np->callback = Flags_ChangeCB;  

			 printk("%<driver.name>: Installed..\n");
                        
		}/

		cleanup /{
			printk("Have a nice day! \n");
		}/
	}

        /*
         * This method will be called whenever the driver is opened
         */
	open /{

		printk("Stupid module has been opened\n");

	}/

        /*
         * This method will be called whenever the driver is closed
         */
	close /{
		printk("Stupid module has been closed\n");
	}/

}

