/@

   This Example shows how signalling from kernel can be done to 
   an user process. The driver module sends the signal with send_sig() 
   to the process that open()ed the device. If send_sig() is called 
   from a different context (this is usually the case from a 
   driver interrupt or timeout handler), the driver has to take care 
   that the process context is stored somewhere.
   In this simple example only one process is stored.

@/

module "Signal" : 
driver "Signal" [ major=34 ] {

        file "Timer" /{
          #include <linux/signal.h>
        	   
          struct timer_list Timer;

          struct task_struct * current_save;

          void SaveTask(){
		current_save = current;
	  }

          void Timer_IRQ(unsigned long unused){
                printk("click! \n");

                send_sig( SIGUSR1 ,current_save,1);

                StartTimer(100);
          }

          void StartTimer(int timeout){
		printk( "start timer\n" );
		init_timer(&Timer); 
		Timer.expires = jiffies + timeout;
		Timer.function = Timer_IRQ;
                add_timer(&Timer); 
          }

	  void StopTimer(){
		del_timer( &Timer );
	  }

        }/

        /*--------------------*/

	core {  
		init /{ 
			printk("init called\n");			
		}/

		cleanup /{
			printk("cleanup called\n");
		}/
	}

	open /{
		printk("open called\n");
		SaveTask();
                StartTimer( 100 );
	}/

	close /{
		printk("close called\n");
		StopTimer();
	}/

}


