/@  

   The 'sysctl' method defines global variables that are accessible
   through the /proc/sys/<driver.name>/<variable>
   interface just by echo or cat. LDDK manages all necessary tasks
   for reading or writing this variables automatically.
   The access modes ( rwx,rwx,rwx ) can be set with the 'access'  
   option.
            
   For handling special tasks as type or value checking,
   an own proc handler can be implemented with the 'handler'
   method directive as shown by this example. 
   The 'strategy' method directive can be used to implement own strategy
   functions (see /usr/src/linux/kernel/sysctl.c for details)

   If the sysctl method is used in a driver the dbgMask variable
   is automatically accessible through the sysctl interface.
   It can just be set to verbose debugging with.

       echo 0x06 >/proc/sys/<driver.name>/dbgMask

   realize that this can be done while the driver is running !
            
@/


module "Sysctl" : 
driver "Sysctl" [ major=34 ] {



	sysctl {

/*---- global system variables, later accessible through /proc/sys/Sysctl/<name> */

		string version = "preliminary test version 0.1"
		int irq   [ access=0444 ] = 5
		int base  [ access=0444 ] = 0x230

                string test [ handler=TestHandler ] = "test the west"

/*---- proc handler */

                handler TestHandler /{ 

			int len;
			char *p, c;

		        printk("Test Handler called :");

	
			if (!table->data || !table->maxlen || !*lenp ||
	    		  (filp->f_pos && !write)) {
			  *lenp = 0;
			  DBGout(); return 0;
			}
	
			if (write) {
                                printk("write");
				len = 0;
				p = buffer;
				while (len < *lenp && 
		       		(c = get_user(p++)) != 0 && c != '\n')
				len++;
			if (len >= table->maxlen)
				len = table->maxlen-1;
				memcpy_fromfs(table->data, buffer, len);
				((char *) table->data)[len] = 0;
				filp->f_pos += *lenp;
			} else {
				printk("read");
				len = strlen(table->data);
				if (len > table->maxlen)
				len = table->maxlen;
				if (len > *lenp)
				len = *lenp;
				if (len)
					memcpy_tofs(buffer, table->data, len);
				if (len < *lenp) {
					put_user('\n', ((char *) buffer) + len);
					len++;
			}
			*lenp = len;
			filp->f_pos += len;
			}
                        printk("\n");
			DBGout(); return 0;

                }/

     }

/*----- the driver cores */

	core {  
		init /{
			printk("init called\n");
		}/

		cleanup /{
			printk("cleanup called\n");
		}/
	}

	open /{
		printk("open called\n");
	}/

	close /{
		printk("close called\n");
	}/

}

