%{
#include <fcntl.h>
#include <rmgr_P.h>

extern int yylineno;
int fd=0;

#define DEVNAME     "/dev/rmgr"
#define DEFAULT_RES "/etc/kres.conf"

#define STRING(a) a,strlen(a)+1

extern void yyerror(char *);
extern void doit();
extern void usage();

char class[120];
char ress[120];
char ress1[120];
%}

%union
{
int ival;
char *sval;
}

%token T_NUMBER T_IDENT T_STRING

%type <ival> T_NUMBER
%type <sval> T_IDENT
%type <sval> T_STRING

%%

line:   /* empty */
	| assignment
	| assignment line
	| error { yyerror("line"); }


assignment:
	'[' T_IDENT ']' { strcpy(class,$2); }
	| T_IDENT ':' T_NUMBER { sprintf(ress,"%s::%s",class,$1); 
				rmgr_SetInt(fd,STRING(ress),$3); 
				TestError(ress);}
	| T_IDENT ':' T_STRING { sprintf(ress,"%s::%s",class,$1); 
                                rmgr_SetString(fd,STRING(ress),STRING($3));
				TestError(ress);}
        | T_IDENT ':' T_IDENT '(' ')' {  sprintf(ress, "%s::%s",class,$1); 
                                         sprintf(ress1,"%s::%s",class,$3); 
                                rmgr_AssignFunc(fd,STRING(ress),STRING(ress1));
				TestError(ress);}
        | error { yyerror("line"); }  

%%

void yyerror(char *s){
printf ("parse error %s line %d\n",s,yylineno);
}

int TestError(char *res){
 extern int errno;
 switch(errno){
	case -ENOTFOUND:
		fprintf(stderr,"Ressource '%s' not found in registry.\n",res);
		break;
	case -ENOPERM:
		fprintf(stderr,"Access to '%s' not permitted.\n",res);
		break;
	case -ENOTYPE:
		fprintf(stderr,"Unable to write '%s' (Type Conflict).\n",res);
		break;
	default:
	  return 0;
	break;
 }
 return 1;
}




char *strsave(char *s)
{
  char *p;
  if(( p = (char *) malloc( strlen(s) +1 )) != NULL)
    strcpy(p,s);
  return(p);

}

#include <sys/param.h>
#include <getopt.h>


main(int argc,char **argv)
{

  int c;
  char devname[MAXPATHLEN];
  char buf[MAXPATHLEN];
  extern FILE *yyin;


  strcpy(devname,DEVNAME);

  while(1){
             int this_option_optind = optind ? optind : 1;
             int option_index = 0;
             static struct option long_options[] =
             {
               {"help",   0, 0, 0},
               {"device", 1, 0, 0},
	       {"delete", 1, 0, 0},
	       {"query",  1, 0, 0},	
	       {"set",    1, 0, 0},	
               {0, 0, 0, 0}
             };

             c = getopt_long (argc, argv, "d:",
                        long_options, &option_index);

	     switch(c){
	     case 0:
	       switch( option_index ){
	       case 0:
		 usage();
		 break;
	       case 1:
		 strcpy( devname ,optarg );
		 break;
	       case 2:
		 strcpy( buf ,optarg );
	         if(( fd = open(devname,O_RDWR)) < 0 ){
                     fprintf(stderr,"Can't open Ressource Manager device '%s'\n",devname);
                     exit(1);
                 }
		 rmgr_Delete(fd,STRING(buf)); 
		 if( TestError(buf) ) 
			exit(1);
		 close(fd); exit(0);	 
		 break;
	       case 3:
		 {
		 int type=0;
		 int aint;
		 char buf2[512];

                 strcpy( buf ,optarg );
	         if(( fd = open(devname,O_RDWR)) < 0 ){
                     fprintf(stderr,"Can't open Ressource Manager device '%s'\n",devname);
                     exit(1);
                 }
		 type = rmgr_GetType(fd,STRING(buf));
		 if( TestError(buf) ) 
			exit(1);
		 switch( type ){
			case RCS_INT:
		        case RCS_IPTR:
			  aint = rmgr_GetInt(fd,STRING(buf));
			  fprintf(stderr," %s = %d \n",buf,aint);
			break;
			  case RCS_STRING:
			    aint = rmgr_GetString(fd,STRING(buf),buf2);
			    printf(" %s = '%s' \n",buf,buf2);
			break;   			  
		 }
		 close(fd); exit(0);
		 }	 
		 break;
	       case 4:
		{
		 int type=0;
		 int aint;
		 char *s;
		 

                 strcpy( buf ,optarg );
		 if( optind < argc ){
			s=argv[optind++];
		 } else {
			fprintf(stderr,"Missing Argument \n");
			exit(1);
		 }
	         if(( fd = open(devname,O_RDWR)) < 0 ){
                     fprintf(stderr,"Can't open Ressource Manager device '%s'\n",devname);
                     exit(1);
                 }
		 type = rmgr_GetType(fd,STRING(buf));
		 if( TestError(buf) ) 
			exit(1);
		 switch(type) {
			case RCS_INT:
			case RCS_IPTR:
			 rmgr_SetInt(fd,STRING(buf),atoi(s));			  
			break;
			case RCS_STRING:
			 rmgr_SetString(fd,STRING(buf),STRING(s));
			break;
			case RCS_FPTR:
			 rmgr_AssignFunc(fd,STRING(buf),STRING(s));
			 TestError(buf);
			break;
		 }
		 if( TestError(buf) ) 
			exit(1);
		 close(fd); exit(0);
		}	
		break;
	       default:
		 exit(1);
		 break;
	       }
              break;
             case 'd':
                 strcat( devname , optarg );
	       break;
             case '?':
               usage();
               break;

             default:
	       /*printf ("?? getopt returned character code 0%o ??0",c);*/
	       break;

	      }

             if (c == -1)
             break;
  }

             if (optind < argc)
             {
               char *s; 
               while (optind < argc) {
		 s = argv[optind++];
		 if(( yyin = fopen( s ,"r" )) == NULL ){
		   printf("Can't open input %s \n",s);
		   exit(1);
		 }  
		 doit(devname);
	       }

             }
             else {
		/* using default */
		 if(( yyin = fopen( DEFAULT_RES ,"r" )) == NULL ){
		   printf("Can't open ressource file %s \n",DEFAULT_RES );
		   exit(1);
		 }  
		 doit(devname);		
	     }

  exit(0);
}

void doit(char *devname){
   char version[30];
   int len;
   extern int errno;

   if(( fd = open(devname,O_RDWR)) < 0 ){
     fprintf(stderr,"Can't open Ressource Manager device '%s'\n",devname);
     exit(1);
   }

   len = rmgr_GetString(fd,STRING("rmgr::version"),version);
   if( errno >= 0 ){
      /*fprintf(stderr,"Rmgr version %s\n",version);*/
      yyparse();
   } else {
      fprintf(stderr,"Device is not the Ressource Manager\n");	
      exit(1);
   }
   close(fd);
}

void usage(){
fprintf(stderr,"\
rmanip -- Kernel Ressource Manipulation Program \n\
          (c)1996 C.Schroeter (clausi@chemie.fu-berlin.de)\n\
Usage:\n\
     rmanip [options] [ressourcefile]\n\
     rmanip command [argument]\n\
Options:\n\
--help,   -?                    : print this message \n\
--device, -d devicename         : specify alternate rmgr device\n\
\n\
Commands:\n\
--query      ressource          : displays a ressource's value\n\
--delete     ressource          : deletes a ressource\n\
--set        ressource value    : sets a ressource to another value\n\
\n");

exit(0);
}
