#include <%<driver.name>_P.h>
#include <linux/mm.h>
#include <linux/sysctl.h>
#include <linux/ctype.h>
#include <%<driver.name>_sysctl.h>

/*
 * This Template implements the SYSCTL basics, and handler/strategy routines
 * Users may implement own routines and hook them up with the 'handler'		
 * and 'strategy' methods of sysctl.
 * 
 *
 */

#define SYSCTL_%<driver.name> 1

/* ----- Prototypes */

extern int %<driver.name>_dointvec(ctl_table *table, int write, struct file *filp,
		  void *buffer, size_t *lenp);

extern int %<driver.name>_dostring(ctl_table *table, int write, struct file *filp,
		  void *buffer, size_t *lenp);


extern int %<driver.name>_sysctl_string(ctl_table *table, int *name, int nlen,
		  void *oldval, size_t *oldlenp,
		  void *newval, size_t newlen, void **context);


/* ----- global variables accessible through /proc/sys/%<driver.name> */

%<driver.sysctl.vars>

/* ----- the sysctl table */

ctl_table %<driver.name>_sysctl_table[] = {
   %<driver.sysctl.table>   
   {0}
};

/* ----- the main directory entry in /proc/sys */

ctl_table %<driver.name>_sys_table[] = {
	    {SYSCTL_%<driver.name>,"%<driver.name>",      NULL, 0, 0555, 
                 &%<driver.name>_sysctl_table},	
	    {0}	
};

/* ----- register and unregister entrys */

struct ctl_table_header *%<driver.name>_systable;

        void register_systables(){
	    %<driver.name>_systable = register_sysctl_table( &%<driver.name>_sys_table, 0 );
        }

        void unregister_systables(){
		unregister_sysctl_table(%<driver.name>_systable);
        }


/* ----- default proc handlers */

int %<driver.name>_dointvec(ctl_table *table, int write, struct file *filp,
		  void *buffer, size_t *lenp)
         {
	int *i, vleft, first=1, len, left, neg, val;
	#define TMPBUFLEN 20
	char buf[TMPBUFLEN], *p;
	
	if (!table->data || !table->maxlen || !*lenp ||
	    (filp->f_pos && !write)) {
		*lenp = 0;
		return 0;
	}
	
	i = (int *) table->data;
	vleft = table->maxlen / sizeof(int);
	left = *lenp;
	
	for (; left && vleft--; i++, first=0) {
		if (write) {
			while (left && isspace(get_user((char *) buffer)))
				left--, ((char *) buffer)++;
			if (!left)
				break;
			neg = 0;
			len = left;
			if (len > TMPBUFLEN-1)
				len = TMPBUFLEN-1;
			memcpy_fromfs(buf, buffer, len);
			buf[len] = 0;
			p = buf;
			if (*p == '-' && left > 1) {
				neg = 1;
				left--, p++;
			}
			if (*p < '0' || *p > '9')
				break;
			val = simple_strtoul(p, &p, 0);
			len = p-buf;
			if ((len < left) && *p && !isspace(*p))
				break;
			if (neg)
				val = -val;
			buffer += len;
			left -= len;
			*i = val;
		} else {
			p = buf;
			if (!first)
				*p++ = '\t';
			sprintf(p, "%d", *i);
			len = strlen(buf);
			if (len > left)
				len = left;
			memcpy_tofs(buffer, buf, len);
			left -= len;
			buffer += len;
		}
	}

	if (!write && !first && left) {
		put_user('\n', (char *) buffer);
		left--, buffer++;
	}
	if (write) {
		p = (char *) buffer;
		while (left && isspace(get_user(p++)))
			left--;
	}
	if (write && first)
		return -EINVAL;
	*lenp -= left;
	filp->f_pos += *lenp;
	return 0;
}


int %<driver.name>_dostring(ctl_table *table, int write, struct file *filp,
		  void *buffer, size_t *lenp)
{
	int len;
	char *p, c;
	
	if (!table->data || !table->maxlen || !*lenp ||
	    (filp->f_pos && !write)) {
		*lenp = 0;
		return 0;
	}
	
	if (write) {
		len = 0;
		p = buffer;
		while (len < *lenp && 
		       (c = get_user(p++)) != 0 && c != '\n')
			len++;
		if (len >= table->maxlen)
			len = table->maxlen-1;
		memcpy_fromfs(table->data, buffer, len);
		((char *) table->data)[len] = 0;
		filp->f_pos += *lenp;
	} else {
		len = strlen(table->data);
		if (len > table->maxlen)
			len = table->maxlen;
		if (len > *lenp)
			len = *lenp;
		if (len)
			memcpy_tofs(buffer, table->data, len);
		if (len < *lenp) {
			put_user('\n', ((char *) buffer) + len);
			len++;
		}
		*lenp = len;
		filp->f_pos += len;
	}
	return 0;
}

/* ----- strategy handlers */

int %<driver.name>_sysctl_string(ctl_table *table, int *name, int nlen,
		  void *oldval, size_t *oldlenp,
		  void *newval, size_t newlen, void **context)
{
	int l, len;
	
	if (!table->data || !table->maxlen) 
		return -ENOTDIR;
	
	if (oldval && oldlenp && get_user(oldlenp)) {
		len = get_user(oldlenp);
		l = strlen(table->data);
		if (len > l) len = l;
		if (len >= table->maxlen)
			len = table->maxlen;
		memcpy_tofs(oldval, table->data, len);
		put_user(0, ((char *) oldval) + len);
		put_user(len, oldlenp);
	}
	if (newval && newlen) {
		len = newlen;
		if (len > table->maxlen)
			len = table->maxlen;
		memcpy_fromfs(table->data, newval, len);
		if (len == table->maxlen)
			len--;
		((char *) table->data)[len] = 0;
	}
	return 0;
}

